#include <JuceHeader.h>

class StandaloneAudioIODeviceCallback : public juce::AudioIODeviceCallback
{
public:
	int64_t timesCalledBack{ 0 };
	bool isRunning{ false };
	juce::String lastErrorMessage {};

	~StandaloneAudioIODeviceCallback() override = default;

	void audioDeviceIOCallback(
		const float** inputChannelData,
		int numInputChannels,
		float** outputChannelData,
		int numOutputChannels,
		int numSamples) override
	{
		timesCalledBack++;
	}

	void audioDeviceAboutToStart(juce::AudioIODevice* device) override
	{
		isRunning = true;
	}

	void audioDeviceStopped() override
	{
		isRunning = false;
	}

	void audioDeviceError(const juce::String& errorMessage) override
	{
		lastErrorMessage = errorMessage;
	}

	void dumpDescription(juce::StringRef subject) const
	{
		DBG(subject);
		DBG("  Times called back: " << timesCalledBack);
		DBG("  Is running:        " << (isRunning ? "yes" : "no"));
		DBG("  Error:             " << lastErrorMessage);
	}
};

class CoreAudioMultiDeviceUsageApplication : public juce::JUCEApplication, private juce::Timer
{
public:
	CoreAudioMultiDeviceUsageApplication()
	{
	}

	const juce::String getApplicationName() override
	{
		return ProjectInfo::projectName;
	}
	const juce::String getApplicationVersion() override
	{
		return ProjectInfo::versionString;
	}
	bool moreThanOneInstanceAllowed() override
	{
		return true;
	}

	void initialise(const juce::String& commandLine) override
	{
		// Create Core Audio backend
		mCoreAudioIODeviceType.reset(juce::AudioIODeviceType::createAudioIODeviceType_CoreAudio());
		jassert(mCoreAudioIODeviceType);

		// Get default audio device name
		mCoreAudioIODeviceType->scanForDevices();
		auto defaultDeviceName =
			mCoreAudioIODeviceType->getDeviceNames()[mCoreAudioIODeviceType->getDefaultDeviceIndex(false)];
		jassert(defaultDeviceName.isNotEmpty());

		DBG("Opening device with name: " << defaultDeviceName);

		// Create audio devices
		mDeviceA.reset(mCoreAudioIODeviceType->createDevice(defaultDeviceName, {}));
		mDeviceB.reset(mCoreAudioIODeviceType->createDevice(defaultDeviceName, {}));

		juce::String errorMessage;
		errorMessage = mDeviceA->open(0, 3, 44100, 128);
		jassert(errorMessage.isEmpty());
		errorMessage = mDeviceB->open(0, 3, 44100, 128);
		jassert(errorMessage.isEmpty());

		mDeviceA->start(&mCallbackA);
		// The next call fails to attach the underlying Core Audio IOProc callback because the pointer given to
		// AudioDeviceStat() is already registered for device A
		mDeviceB->start(&mCallbackB);

		// Now only device A will get called back

		startTimer(2000);
	}

	void shutdown() override
	{
	}

	//==============================================================================
	void systemRequestedQuit() override
	{
		quit();
	}

	void anotherInstanceStarted(const juce::String& commandLine) override
	{
	}

private:
	std::unique_ptr<juce::AudioIODeviceType> mCoreAudioIODeviceType;
	std::unique_ptr<juce::AudioIODevice> mDeviceA;
	std::unique_ptr<juce::AudioIODevice> mDeviceB;
	StandaloneAudioIODeviceCallback mCallbackA;
	StandaloneAudioIODeviceCallback mCallbackB;

	void timerCallback() override
	{
		// jassert(mCallbackA.timesCalledBack == mCallbackB.timesCalledBack); // This assert will be hit when uncommented

		mCallbackA.dumpDescription("Device A");
		mCallbackB.dumpDescription("Device B");

		DBG("");
	}
};

//==============================================================================
// This macro generates the main() routine that launches the app.
START_JUCE_APPLICATION (CoreAudioMultiDeviceUsageApplication)
